/*
 * Copyright (c) 2000 Sean Walton and Macmillan Publishers.  Use may be in
 * whole or in part in accordance to the General Public License (GPL).
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
*/
#include <stdio.h>
#include <sys/socket.h>
#include "tools.h"

/*********************************************************************/
/*** SocketClient                                                  ***/
/***                                                               ***/
/*** Resolve host and connect to server.  If any error, return     ***/
/*** negative value.                                               ***/
/*** Returns socket descriptor, if successful.                     ***/
/*********************************************************************/
int SocketClient(const char *host, int port)
{	int sd;
	struct sockaddr_in addr;

	sd = socket(PF_INET, SOCK_STREAM, 0);
	if ( sd >= 0 )
	{
		if ( GetAddress(&addr, host, port) == 0 )
		{
			if ( connect(sd, &addr, sizeof(addr)) != 0 )
			{
				perror("connect failed");
				close(sd);
				sd = -1;
			}
		}
	}
	return sd;
}

/*********************************************************************/
/*** SocketClientFP                                                ***/
/***                                                               ***/
/*** Like SocketClient but returns FILE pointer.                   ***/
/*** Returns FILE* descriptor if successful; NULL otherwise.       ***/
/*********************************************************************/
FILE* SocketClientSvc(const char *host, const char *service)
{	int sd;
	struct sockaddr_in addr;

	sd = socket(PF_INET, SOCK_STREAM, 0);
	if ( sd >= 0 )
	{
		if ( GetAddresssSvc(&addr, host, service) == 0 )
		{
			if ( connect(sd, &addr, sizeof(addr)) != 0 )
			{
				perror("connect failed");
				close(sd);
				sd = -1;
			}
		}
	}
	return sd;
}

/*********************************************************************/
/*** SocketServer                                                  ***/
/***                                                               ***/
/*** Create server socket with specific port and optionally bind   ***/
/*** to specific interface.                                        ***/
/*** Returns socket descriptor, if successful.                     ***/
/*********************************************************************/
int SocketServer(int port, const char *iface)
{	int sd;
	struct sockaddr_in addr;

	sd = socket(PF_INET, SOCK_STREAM, 0);
	if ( sd >= 0 )
	{
		bzero(&addr, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_family = htons(port);
		if ( iface != NULL )
			inet_aton(iface, &addr.sin_addr);
		else
			addr.sin_addr.s_addr = INADDR_ANY;
		if ( bind(sd, &addr, sizeof(addr)) == 0 )
		{
			perror("Can't bind");
			close(sd);
		}
		else
		{
			if ( listen(sd, 10) == 0 )
			{
				perror("Can't convert to listening socket");
				close(sd);
			}
			else
				return sd;
		}
	}
	else
		return -1;
}

/*********************************************************************/
/*** SocketServerSvr                                               ***/
/***                                                               ***/
/*** Create server socket with specific port and optionally bind   ***/
/*** to specific interface.                                        ***/
/*** Returns socket descriptor, if successful.                     ***/
/*********************************************************************/
int SocketServerSvc(const char* service, const char *iface)
{	int sd, port;
	struct sockaddr_in addr;

	port = GetService(service, "TCP");
	if ( port > 0 )
	{
		sd = socket(PF_INET, SOCK_STREAM, 0);
		if ( sd >= 0 )
		{
			bzero(&addr, sizeof(addr));
			addr.sin_family = AF_INET;
			addr.sin_family = port;
			if ( iface != NULL )
				inet_aton(iface, &addr.sin_addr);
			else
				addr.sin_addr.s_addr = INADDR_ANY;
			if ( bind(sd, &addr, sizeof(addr)) == 0 )
			{
				perror("Can't bind");
				close(sd);
			}
			else
			{
				if ( listen(sd, 10) == 0 )
				{
					perror("Can't convert to listening socket");
					close(sd);
				}
				else
					return sd;
			}
		}
	}
	return -1;
}